#define PYGII_MODULE
#include "pygii.h"

#define INPUT_FUNC(s, c) PyCFunction_New(&Input_methods[c], (PyObject*)s)

static PyObject*
Input_split(Input * self, PyObject* args) {
	gii_input_t inp;
	Input *ret;
	long orig;
	int r;
	
	if(!PyArg_ParseTuple(args, "i", &orig)) {
		PyErr_SetString(PyExc_TypeError, "origin required");
		return NULL;
	}
	
	r = giiSplitInputs(self->inp, &inp, (uint32)orig, (uint32)0);
	
	if(r == GGI_ENOTFOUND) {
		PyErr_SetString(PyExc_ValueError, "device not found");
		return NULL;
	}

	ret = PyObject_New(Input, &InputType);
	if(!ret) return NULL;
	ret->inp = inp;
	return (PyObject*)ret;
}

static PyObject*
Input_join(Input * self, PyObject* args) {
	PyObject * inp;
	
	if(!PyArg_ParseTuple(args, "O", &inp)) {
		PyErr_SetString(PyExc_TypeError, "Input required");
		return NULL;
	}
	
	ASSERT_INPUT(inp,   NULL);
	ASSERT_VALID_INPUT(self, NULL);

	self->inp = giiJoinInputs(self->inp,((Input*)inp)->inp);
	
	((Input*)inp)->inp = NULL; /* invalidate the object */
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Input_read(Input *self, PyObject *args, PyObject *kwds) {
	int mask;
	gii_event e;
	gii_event *ev;
	PyObject * d;
	static char *kwlist[]={ "mask", "event", NULL };
	
	mask = emAll; /* by default */
	ev   = &e;
	d    = NULL;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|iO",
					kwlist, &mask, &d)) {
		return NULL;
	}

	if (d) {
		ASSERT_EVENT(d, NULL);
		ev=&(((Event*)(d))->ev);
	}
	
	if(giiEventsQueued(self->inp,mask)) {
		giiEventRead(self->inp, ev, mask);
		if (d==NULL)
			return pygii_newEventFrom(e);
		else
			/* XXX do we need to drop a reference? */
			return d;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Input_poll(Input *self, PyObject*args, PyObject *kwds) {
	struct timeval t;
	struct timeval * tv;
	double f=-1.0;
	gii_event_mask mask;
	static char *kwlist[]={ "timeout", "mask", NULL };
	
	mask=emAll;
	t.tv_sec=0;
	t.tv_usec=0;
	tv = NULL;
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|di",
					kwlist, &f, &mask)) {
		return NULL;
	}
	
	if (f>=0) {
		FLOAT_TO_TS(&t,f);
		tv=&t;
	}
	mask=giiEventPoll(self->inp, mask,tv);
	if (mask==0) {
		pygii_raiseTimeout();
		return NULL;
	}
	if (tv==NULL) /* no timeout */ 
		return PyInt_FromLong((long)mask);	
	return Py_BuildValue("id", mask, TS_TO_FLOAT(&t));
}

static PyObject*
Input_queued(Input *self, PyObject*args, PyObject *kwds) {
	gii_event_mask mask;
	static char *kwlist[]={ "mask", NULL };
	
	mask=emAll; /* by default */
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|i",
					kwlist, &mask))
		return NULL;
	return PyInt_FromLong((long)giiEventsQueued(self->inp, mask));
}

static PyObject*
Input_send(Input *self, PyObject* args) {
	PyObject * e;
	
	if(!PyArg_ParseTuple(args, "O", &e)) {
		PyErr_SetString(PyExc_TypeError, "Event required");
		return NULL;
	}
	
	ASSERT_EVENT(e,NULL);
	if(giiEventSend(self->inp, &(((Event*)e)->ev))<0) {
		PyErr_SetString(PyExc_SystemError, "giiEventSend failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}


enum {
	INPUT_JOIN = 0,
	INPUT_SPLIT,
	INPUT_READ,
	INPUT_POLL,
	INPUT_QUEUED,
	INPUT_SEND
};

static struct PyMethodDef Input_methods[] = {
	/* ---------- Input management ----------------*/

	{ "join",    (PyCFunction)Input_join,   METH_VARARGS, (char*)NULL  },
	{ "split",   (PyCFunction)Input_split,  METH_VARARGS, (char*)NULL },
	
	/*------------- Events ------------------------*/

	{ "read",  (PyCFunction)Input_read,   METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL },
	{ "poll",  (PyCFunction)Input_poll,   METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL },
	{ "queued",(PyCFunction)Input_queued, METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL},
	{ "send",  (PyCFunction)Input_send,   METH_VARARGS,
	  (char*)NULL},

	/* ---------- end -----------*/
	
	{ NULL, NULL, 0, (char*)NULL }
};

static PyObject*
Input_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
	Input *self;
	
	self = (Input*)type->tp_alloc(type, 0);
	if (self!=NULL) {
		self->inp = NULL;
	}
	return (PyObject*)self;
}

static int
Input_init(Input * self, PyObject *args, PyObject *kwds) {
	char *a;

	a = NULL;
	if(giiInit()<0) {
		PyErr_SetString(PyExc_SystemError,"giiInit failed");
		return -1;
	}
	if(!PyArg_ParseTuple(args, "|s", &a)) {
		PyErr_SetString(PyExc_TypeError,"string required");
		return -1;
	}
	self->inp = giiOpen(a);
	if (self->inp == NULL) {
		PyErr_SetString(PyExc_SystemError,"giiOpen failed");
		return -1;
	}
	
       	return 0;
}

static void
Input_dealloc(Input *self) {
	if(self->inp)
		giiClose(self->inp);
	giiExit();
	self->ob_type->tp_free((PyObject*)self);
}

static int
Input_setattr(Input *self, char *name, PyObject *value) {
	int mask;

	ASSERT_VALID_INPUT(self, -1);
	
	switch(name[0]) {
	case 'm':
		if(!strcmp(name,"mask")) {
			if(pygii_getInt(value, &mask)<0)
				return -1;
			giiSetEventMask(self->inp, mask);
			return 0;
		}
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

static PyObject*
Input_getattr(Input *self, char *name) {
	ASSERT_VALID_INPUT(self, NULL);
	switch(name[0]) {
	case 'b':
		break;
	case 'd':
		if(!strcmp(name,"devices"))
			return pygii_getDeviceInfo(self->inp);
		break;
	case 'j':
		if(!strcmp(name,"join"))
			return INPUT_FUNC(self, INPUT_JOIN);
		break;
	case 'm':
		if(!strcmp(name,"mask"))
			return PyInt_FromLong((long)giiGetEventMask(self->inp));
		break;
	case 'p':
		if(!strcmp(name,"poll"))
			return INPUT_FUNC(self, INPUT_POLL);
		break;
	case 'q':
		if(!strcmp(name,"queued"))
			return INPUT_FUNC(self, INPUT_QUEUED);
		break;
	case 'r':
		if(!strcmp(name,"read"))
			return INPUT_FUNC(self, INPUT_READ);
		break;
	case 's':
		if(!strcmp(name,"send"))
			return INPUT_FUNC(self, INPUT_SEND);
		if(!strcmp(name,"split"))
			return INPUT_FUNC(self, INPUT_SPLIT);
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static PyObject*
Input_str(Input * self) {
	return PyString_FromFormat("<gii_input_t at %p>",self->inp);
}

static PyObject*
Input_repr(Input * self) {
	return PyString_FromString("Input()");
}


PyTypeObject InputType = {
	PyObject_HEAD_INIT(NULL)
	0,                          /* ob_size         */
	"GGI._giimodule.Input",     /* tp_name         */
	sizeof(Input),              /* tp_basicsize    */
	0,                          /* tp_itemsize     */
	(destructor)Input_dealloc,  /* tp_dealloc      */
	0,                          /* tp_print        */
	(getattrfunc)Input_getattr, /* tp_getattr      */
	(setattrfunc)Input_setattr, /* tp_setattr      */
	0,                          /* tp_compare      */
	(reprfunc)Input_repr,       /* tp_repr         */
	0,                          /* tp_as_number    */
	0,                          /* tp_as_sequence  */
	0,                          /* tp_as_mapping   */
	0,                          /* tp_hash         */
	0,                          /* tp_call         */
	(reprfunc)Input_str,        /* tp_str          */
	0,                          /* tp_getattro     */
	0,                          /* tp_setattro     */
	0,                          /* tp_as_buffer    */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,     /*tp_flags*/
	"Input objects",            /* tp_doc          */
	
	/* 2.0 */
	0,		            /* tp_traverse     */
	0,		            /* tp_clear        */
	/* 2.1 */
	0,		            /* tp_richcompare  */
	0,		            /* tp_weaklistoffset */
	/* 2.2 */
	0,		            /* tp_iter         */
	0,		            /* tp_iternext     */
	0,                          /* tp_methods      */
	0,                          /* tp_members      */
	0,                          /* tp_getset       */
	0,                          /* tp_base         */
	0,                          /* tp_dict         */
	0,                          /* tp_descr_get    */
	0,                          /* tp_descr_set    */
	0,                          /* tp_dictoffset   */
	(initproc)Input_init,       /* tp_init         */
	0,                          /* tp_alloc        */
	Input_new,                  /* tp_new          */
	0,                          /* tp_is_gc        */
	0,                          /* tp_bases        */
        0,                          /* tp_mro          */
	0,                          /* tp_cache        */
	0,                          /* tp_subclasses   */
	0,                          /* tp_weaklist     */
	0                           /* tp_del          */
};
